from transformers import AutoTokenizer

def load_default_tokenizer():
    tokenizer = AutoTokenizer.from_pretrained("meta-llama/Llama-2-7b-hf")
    tokenizer.padding_side = "right" # Fix weird overflow issue with fp16 training

    return tokenizer


import logging
from colorama import init, Fore, Style

# Initialize colorama
init(autoreset=True)

# Create a custom formatter to add colors
class CustomFormatter(logging.Formatter):
    """Logging Formatter to add colors"""

    grey = Style.DIM + Fore.WHITE
    yellow = Style.BRIGHT + Fore.YELLOW
    green = Style.BRIGHT + Fore.GREEN
    format = "%(asctime)s - %(levelname)s - %(message)s"

    FORMATS = {
        logging.DEBUG: grey + format,
        logging.INFO: green + format,
        logging.WARNING: yellow + format,
        logging.ERROR: Fore.RED + format,
        logging.CRITICAL: Fore.RED + Style.BRIGHT + format
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)

# Function to set up and return a logger
def get_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    if not logger.handlers:
        # Create console handler with a higher log level
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)

        # Create formatter and add it to the handler
        ch.setFormatter(CustomFormatter())

        # Add the handler to the logger
        logger.addHandler(ch)

    return logger


import requests

from flex_ai.common.enums import DatasetType
BASE_URL = "https://api.getflex.ai"
# BASE_URL = "http://localhost:8080"
# send api key in the header
def generate_dataset_upload_urls(api_key:str, dataset_id:str):
    url = f"{BASE_URL}/v1/datasets/generate_upload_urls"
    headers = {
        "Authorization": f"Bearer {api_key}",
        "Content-Type": "application/json"
    }
    payload = {"id": dataset_id}

    response = requests.post(url, json=payload, headers=headers)
    if response.status_code != 200:
        raise Exception(response.json()["detail"])
    data = response.json()
    return data["train_upload_url"], data["eval_upload_url"]

def create_dataset(api_key:str, dataset_id: str, name:str, train_rows_count:int, eval_rows_count:int, max_tokens:int, total_tokens:int, type: DatasetType):
    url = f"{BASE_URL}/v1/datasets/create_dataset"
    headers = {
        "Authorization": f"Bearer {api_key}",
        "Content-Type": "application/json"
    }
    payload = {"id": dataset_id, "name": name, "train_rows_count": train_rows_count, "eval_rows_count": eval_rows_count, "max_tokens": max_tokens, "total_tokens": total_tokens, "type": type.value}

    response = requests.post(url, json=payload, headers=headers)

    if response.status_code != 200:
        raise Exception(response.json()["detail"])
    
    data = response.json()
    return data[0]


import json
import shutil
import os

# Read the .jsonl file and parse the data
def read_jsonl(file_path: str):
    with open(file_path, 'r') as file:
        return [json.loads(line) for line in file]
    
def resolve_path(id: str, file_path: str):
    if "{{id}}" in file_path:
        file_path = file_path.replace("{{id}}", id)
    return file_path

def copy_files(file_path: str, source_path: str):
    # Create the directory path if it doesn't exist
    os.makedirs(file_path, exist_ok=True)
    
    # Copy files from the source path to the destination path
    for file_name in os.listdir(source_path):
        source_file = os.path.join(source_path, file_name)
        destination_file = os.path.join(file_path, file_name)
        shutil.copy2(source_file, destination_file)


from typing import Any, Dict, Optional
import requests

from flex_ai.common.classes import EarlyStoppingConfig, LoraConfig
from flex_ai.common.enums import DatasetType

# BASE_URL = "https://api.getflex.ai"
BASE_URL = "http://localhost:8080"
# send api key in the header
def generate_dataset_upload_urls(api_key:str, dataset_id:str):
    url = f"{BASE_URL}/v1/datasets/generate_upload_urls"
    headers = {
        "Authorization": f"Bearer {api_key}",
        "Content-Type": "application/json"
    }
    payload = {"id": dataset_id}

    response = requests.post(url, json=payload, headers=headers)
    if response.status_code != 200:
        raise Exception(response.json()["detail"])
    data = response.json()
    return data["train_upload_url"], data["eval_upload_url"]

def create_finetune(api_key:str, name:str, dataset_id: str, 
                        model: str, n_epochs: int,
                        train_with_lora: bool,
                        batch_size: Optional[int] = None, learning_rate: Optional[float] = None,
                        n_checkpoints_and_evaluations_per_epoch: Optional[int] = None,
                        save_only_best_checkpoint: bool = False,
                        lora_config: Optional[LoraConfig] | None = None,
                        wandb_key: Optional[str] = None,
                        early_stopping_config: Optional[EarlyStoppingConfig] | None = None):
    
    url = f"{BASE_URL}/v1/fine_tunes/create_finetune"
    headers = {
        "Authorization": f"Bearer {api_key}",
        "Content-Type": "application/json"
    }
    # Construct the request data dynamically
    payload: Dict[str, Any] = {
        "name": name,
        "dataset_id": dataset_id,
        "model": model,
        "train_with_lora": train_with_lora
    }
    
    if n_epochs is not None:
        payload["n_epochs"] = n_epochs
    if batch_size is not None:
        payload["batch_size"] = batch_size
    if wandb_key is not None:
        payload["wandb_key"] = wandb_key
    if learning_rate is not None:
        payload["learning_rate"] = learning_rate
    if n_checkpoints_and_evaluations_per_epoch is not None:
        payload["n_checkpoints_and_evaluations_per_epoch"] = n_checkpoints_and_evaluations_per_epoch
    if save_only_best_checkpoint is not None:
        payload["save_only_best_checkpoint"] = save_only_best_checkpoint
    if lora_config is not None:
        payload["lora_config"] = lora_config  # Convert Pydantic model to dictionary
    if early_stopping_config is not None:
        payload["early_stopping_config"] = early_stopping_config  # Convert Pydantic model to dictionary

    response = requests.post(url, json=payload, headers=headers)

    if response.status_code != 200:
        raise Exception(response.json()["detail"])
    data = response.json()
    return data[0]


from flex_ai.common import enums
from flex_ai.utils.collators import ensure_spaces_between_chat_completions_separators
from flex_ai.utils.datasets import convert_instruction_dataset_to_chat_dataset
from flex_ai.data_loaders.files import read_jsonl
from datasets import Dataset
import pandas as pd
from flex_ai.common.logger import get_logger
from transformers import PreTrainedTokenizerBase

logger = get_logger(__name__)


def validate_dataset(train_path:str, eval_path:str | None, type: enums.DatasetType, tokenizer: PreTrainedTokenizerBase):
    train_data = read_jsonl(train_path)
    eval_data = read_jsonl(eval_path)

    if type == enums.DatasetType.CHAT:
        train_data = ensure_spaces_between_chat_completions_separators(train_data)
        eval_data = ensure_spaces_between_chat_completions_separators(eval_data)

        transformed_train_data = [
            {"text": tokenizer.apply_chat_template(chat, tokenize=False)}
            for chat in train_data
        ]
        transformed_eval_data = [
            {"text": tokenizer.apply_chat_template(chat, tokenize=False)}
            for chat in eval_data
        ]
        train_dataset = Dataset.from_pandas(pd.DataFrame(transformed_train_data)).shuffle(seed=42)
        eval_dataset = Dataset.from_pandas(pd.DataFrame(transformed_eval_data)).shuffle(seed=42)
    elif type == enums.DatasetType.INSTRUCTION:
        train_data, eval_data = convert_instruction_dataset_to_chat_dataset(train_data, eval_data)
        train_data = ensure_spaces_between_chat_completions_separators(train_data)
        eval_data = ensure_spaces_between_chat_completions_separators(eval_data)

        transformed_train_data = [
            {"text": tokenizer.apply_chat_template(chat, tokenize=False)}
            for chat in train_data
        ]
        transformed_eval_data = [
            {"text": tokenizer.apply_chat_template(chat, tokenize=False)}
            for chat in eval_data
        ]
        train_dataset = Dataset.from_pandas(pd.DataFrame(transformed_train_data)).shuffle(seed=42)
        eval_dataset = Dataset.from_pandas(pd.DataFrame(transformed_eval_data)).shuffle(seed=42)
    elif type == enums.DatasetType.TEXT:
        train_dataset = Dataset.from_pandas(pd.DataFrame(train_data)).shuffle(seed=42)
        eval_dataset = Dataset.from_pandas(pd.DataFrame(eval_data)).shuffle(seed=42)

    return train_dataset, eval_dataset

def _log_transformed_dpo_examples(train_data, eval_data):
        logger.info("Train dataset chat example after model template for DPO:")
        logger.info("prompt:")
        logger.info(train_data[0]["prompt"])
        logger.info("")
        logger.info("chosen:")
        logger.info(train_data[0]["chosen"])
        logger.info("")
        logger.info("rejected:")
        logger.info(train_data[0]["rejected"])
        logger.info("")
        logger.info("Eval dataset chat example after model template for DPO:")
        logger.info("prompt:")
        logger.info(eval_data[0]["prompt"])
        logger.info("")
        logger.info("chosen:")
        logger.info(eval_data[0]["chosen"])
        logger.info("")
        logger.info("rejected:")
        logger.info(eval_data[0]["rejected"])
        logger.info("")

def _log_transformed_sft_examples(train_data, eval_data):
        logger.info("Train dataset chat example after model template:")
        logger.info(train_data[0]["text"])
        logger.info("")
        logger.info("Eval dataset chat example after model template:")
        logger.info(eval_data[0]["text"])
        logger.info("")


import matplotlib.pyplot as plt

def plot_token_length_distribution(token_lengths, title):
    plt.figure(figsize=(10, 6))
    plt.hist(token_lengths, bins=50, color='blue', alpha=0.7)
    plt.title(title)
    plt.xlabel('Token Length')
    plt.ylabel('Frequency')
    plt.grid(True)
    plt.show()

def generate_report(max_seq_len_train, max_seq_len_eval):
    report = "Dataset is Valid for training:\n"
    report += "===================\n"
    report += f"Max sequence length in train dataset: {max_seq_len_train}\n"
    if max_seq_len_eval is not None:
        report += f"Max sequence length in eval dataset: {max_seq_len_eval}\n"
    else:
        report += "No evaluation dataset provided.\n"

    if max_seq_len_eval and max_seq_len_eval > max_seq_len_train:
        report += "Warning: The evaluation dataset has a longer sequence than the training dataset.\n"
    report += "\nConsiderations:\n"
    report += "- Check if a single example is causing a very high max token length.\n"
    report += "- Adjust context size if necessary.\n"

    return report


from flex_ai.common.logger import get_logger

logger = get_logger(__name__)

# This special function is activated only when train on completion_only is enabled.
# When you train only on completions , you have the make sure the separators tokens [/INST] and [INST],
# have space from both sides , so that we can detect when the completion starts and that the tokenizer wont combine
# their tokens and other tokens. In huggingface prompt default templates - they dont put a space before the beginning of the completions,
# so I do it here manually 
def ensure_spaces_between_chat_completions_separators(chat_data):
    # Iterate through each item in the array
    for conversation in chat_data:
        for answer in conversation:
            # Check if the role is "assistant"
            if answer["role"] == "assistant":
                # Ensure the content starts with a space
                if not answer["content"].startswith(" "):
                    answer["content"] = " " + answer["content"]
    return chat_data

def debug_completion_only_collator(cfg, dataset, tokenizer, collator):
    text_sample = dataset[0]["text"]
    tokenized_text_sample = tokenizer(
            text_sample,
            truncation=True,
            padding=False,
            max_length=cfg.max_seq_length,
            return_overflowing_tokens=False,
            return_length=False,
            add_special_tokens=True
        )
    
    output = collator([tokenized_text_sample])
    input_ids = output["input_ids"]
    labels = output["labels"]

    # For each example in your batch
    for i in range(input_ids.size(0)):
        segments = []
        train_segment_ids = []
        not_train_segment_ids = []

        # Iterate through each token in the example
        for token_id, label in zip(input_ids[i], labels[i]):
            if label.item() == -100:
                if train_segment_ids:
                    train_segment = tokenizer.decode(train_segment_ids, skip_special_tokens=True, clean_up_tokenization_spaces=True)
                    segments.append(f"***{train_segment}***")
                    train_segment_ids = []
                not_train_segment_ids.append(token_id)
            else:
                if not_train_segment_ids:
                    not_train_segment = tokenizer.decode(not_train_segment_ids, skip_special_tokens=True, clean_up_tokenization_spaces=True)
                    segments.append(not_train_segment)
                    not_train_segment_ids = []
                train_segment_ids.append(token_id)

        # Add any remaining segments
        if not_train_segment_ids:
            not_train_segment = tokenizer.decode(not_train_segment_ids, skip_special_tokens=True, clean_up_tokenization_spaces=True)
            segments.append(not_train_segment)
        if train_segment_ids:
            train_segment = tokenizer.decode(train_segment_ids, skip_special_tokens=True, clean_up_tokenization_spaces=True)
            segments.append(f"***{train_segment}***")

        # Combine and print the full text with markers
        full_text_with_markers = "".join(segments)
        logger.info("DataCollatorForCompletionOnlyLM will train only text inside the ***    ***:")
        logger.info(full_text_with_markers)
        logger.info("")



from torch.utils.data import Dataset
from transformers import PreTrainedTokenizerBase

def _instruction_row_to_chat_format(example):
    if "system" in example and example["system"] is not None:
        converted_data = [{ "role": "system",  "content": example["system"]}, { "role": "user",  "content": example["instruction"] },{"role": "assistant", "content": example["output"]}]
    else:
        converted_data = [{ "role": "user",  "content": example["instruction"]},{"role": "assistant", "content": example["output"]}]
    return converted_data

def convert_instruction_dataset_to_chat_dataset(train_data: list[any], eval_data: list[any] ):
    converted_train_data = []
    converted_eval_data = []

    for item in train_data:
        converted_data = _instruction_row_to_chat_format(item)
        converted_train_data.append(converted_data)
    
    for item in eval_data:
        converted_data = _instruction_row_to_chat_format(item)
        converted_eval_data.append(converted_data)
    
    return converted_train_data, converted_eval_data

def convert_instruction_dpo_dataset_to_chat_template(train_data: list[any], eval_data: list[any], tokenizer: PreTrainedTokenizerBase):
    converted_train_data = []
    converted_eval_data = []

    for item in train_data:
        converted_data = _dpo_instruction_row_to_chat_format(item, tokenizer)
        converted_train_data.append(converted_data)
    5
    for item in eval_data:
        converted_data = _dpo_instruction_row_to_chat_format(item, tokenizer)
        converted_eval_data.append(converted_data)
    
    return converted_train_data, converted_eval_data

def convert_chat_dpo_dataset_to_chat_template(train_data: list[any], eval_data: list[any], tokenizer: PreTrainedTokenizerBase):
    converted_train_data = []
    converted_eval_data = []

    for item in train_data:
        converted_data = _dpo_chat_row_to_chat_format(item, tokenizer)
        converted_train_data.append(converted_data)
    5
    for item in eval_data:
        converted_data = _dpo_chat_row_to_chat_format(item, tokenizer)
        converted_eval_data.append(converted_data)
    
    return converted_train_data, converted_eval_data

def _dpo_chat_row_to_chat_format(example, tokenizer: PreTrainedTokenizerBase):
    prompt = tokenizer.apply_chat_template(example["chat"], tokenize=False, add_generation_prompt=True)

    # Format chosen answer
    chosen = example['chosen'] + tokenizer.special_tokens_map['eos_token']

    # Format rejected answer
    rejected = example['rejected'] + tokenizer.special_tokens_map['eos_token']

    return {
        "prompt": prompt,
        "chosen": chosen,
        "rejected": rejected,
    }

def _dpo_instruction_row_to_chat_format(example, tokenizer: PreTrainedTokenizerBase):
    # Format system
    if 'system' in example and example['system'] != "":
        message = {"role": "system", "content": example['system']}
        system = tokenizer.apply_chat_template([message], tokenize=False)
    else:
        system = ""

    # Format instruction
    message = {"role": "user", "content": example['instruction']}
    prompt = tokenizer.apply_chat_template([message], tokenize=False, add_generation_prompt=True)

    # Format chosen answer
    chosen = example['chosen'] + tokenizer.special_tokens_map['eos_token']

    # Format rejected answer
    rejected = example['rejected'] + tokenizer.special_tokens_map['eos_token']

    return {
        "prompt": system + prompt,
        "chosen": chosen,
        "rejected": rejected,
    }


class TextToJsonDataset(Dataset):
    def __init__(self, df, *, tokenizer, in_col='clean_text', out_col='output'):
        self.df = df
        self.in_col = in_col
        self.out_col = out_col
        self.tokenizer = tokenizer
        self.num_rows = len(df)

    def __len__(self):
        return len(self.df)

    def __getitem__(self, idx):
        return {"input_ids":self.tokenizer(self.df.iloc[idx][self.in_col])["input_ids"], "labels":self.tokenizer(self.df.iloc[idx][self.out_col])["input_ids"]}



# python setup.py sdist bdist_wheel
# twine check dist/*
# twine upload dist/*

from setuptools import setup, find_packages

def parse_requirements(filename):
    with open(filename, 'r') as file:
        return file.read().splitlines()

setup(
    name="flex_ai",
    version="0.22",
    packages=find_packages(),
    install_requires=parse_requirements('requirements.txt'),
    author="Ariel Cohen",
    author_email="ariel042cohen@gmail.com",
    description="Flex AI client library",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/arielcohen4/flex_ai",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from pydantic import BaseModel

class LoraConfig(BaseModel):
    lora_r: int
    lora_alpha: int
    lora_dropout: float

class EarlyStoppingConfig(BaseModel):
    patience: int
    threshold: float

class CreateFinetuneRequest(BaseModel):
    dataset_id: str
    model: str
    n_epochs: int
    batch_size: int
    learning_rate: float
    n_checkpoints_and_evaluations_per_epoch: int
    save_only_best_checkpoint: bool
    train_with_lora: bool
    lora_config: LoraConfig
    early_stopping_config: EarlyStoppingConfig


from typing import Optional
import requests
import os
from flex_ai.api.datasets import create_dataset, generate_dataset_upload_urls
from flex_ai.api.fine_tunes import create_finetune
from flex_ai.common import enums
from flex_ai.common.classes import EarlyStoppingConfig, LoraConfig
from flex_ai.data_loaders.loaders import validate_dataset
from flex_ai.common.logger import get_logger
import uuid

logger = get_logger(__name__)

from flex_ai.utils.tokenizers import load_default_tokenizer
from flex_ai.utils.visualize import generate_report

class FlexAI:
    def __init__(self, api_key=None):
        if api_key is None:
            api_key = os.environ.get("FLEX_AI_API_KEY")
        if not api_key:
            raise ValueError("API key must be provided")
        self.api_key = api_key

    def validate_dataset(self, train_path:str, eval_path:str | None, type: enums.DatasetType):
        tokenizer = load_default_tokenizer()
        train_dataset, eval_dataset = validate_dataset(train_path, eval_path, type, tokenizer)

        def tokenize_text(examples):
            return {"num_tokens": [len(tokens) for tokens in tokenizer(examples["text"])["input_ids"]]}
    
        # Apply tokenization to the dataset and compute the maximum token size
        train_dataset_with_tokens = train_dataset.map(tokenize_text, batched=True, num_proc=1)
        max_seq_len_train = max(train_dataset_with_tokens["num_tokens"])

        if eval_dataset:
            eval_dataset_with_tokens = eval_dataset.map(tokenize_text, batched=True, num_proc=1)
            max_seq_len_eval = max(eval_dataset_with_tokens["num_tokens"])
        else:
            max_seq_len_eval = None

        # Print a report
        report = generate_report(max_seq_len_train, max_seq_len_eval)
        logger.info(report)
        
        return f"Using API key: {self.api_key}"
    

    def create_dataset(self, name:str, train_path:str, eval_path:str | None, type: enums.DatasetType):
        tokenizer = load_default_tokenizer()
        train_dataset, eval_dataset = validate_dataset(train_path, eval_path, type, tokenizer)

        def tokenize_text(examples):
            return {"num_tokens": [len(tokens) for tokens in tokenizer(examples["text"])["input_ids"]]}
    
        # Apply tokenization to the dataset and compute the maximum token size
        train_dataset_with_tokens = train_dataset.map(tokenize_text, batched=True, num_proc=1)
        max_seq_len_train = max(train_dataset_with_tokens["num_tokens"])
        total_train_tokens = sum(train_dataset_with_tokens["num_tokens"])

        if eval_dataset:
            eval_dataset_with_tokens = eval_dataset.map(tokenize_text, batched=True, num_proc=1)
            max_seq_len_eval = max(eval_dataset_with_tokens["num_tokens"])
        else:
            max_seq_len_eval = None

        dataset_id = str(uuid.uuid4())
        # upload the train_path and eval_path to the server
        train_upload_url, eval_upload_url = generate_dataset_upload_urls(self.api_key, dataset_id)
        
        # Upload the train dataset file to the server using the pre-signed URL
        with open(train_path, 'rb') as f:
            response = requests.put(train_upload_url, data=f)
            if response.status_code == 200:
                print("Train dataset uploaded successfully.")
            else:
                upload_success = False
                print(f"Failed to upload train dataset. Status code: {response.status_code}")
                return

        if eval_path:
            with open(eval_path, 'rb') as f:
                response = requests.put(eval_upload_url, data=f)
                if response.status_code == 200:
                    print("Eval dataset uploaded successfully.")
                else:
                    print(f"Failed to upload eval dataset. Status code: {response.status_code}")
                    return

        new_dataset = create_dataset(self.api_key, dataset_id, name, len(train_dataset), len(eval_dataset) if eval_dataset else None, max_seq_len_train, total_train_tokens , type)
        print("New Dataset created successfully.")
        print(new_dataset)
        
        return new_dataset
    

    def create_finetune(self, 
                        name:str, dataset_id: str, 
                        model: str, n_epochs: int,
                        train_with_lora: bool,
                        batch_size: Optional[int] = None, learning_rate: Optional[float] = None,
                        wandb_key: Optional[str] = None,
                        n_checkpoints_and_evaluations_per_epoch: Optional[int] = None,
                        save_only_best_checkpoint: bool = False,
                        lora_config: Optional[LoraConfig] | None = None,
                        early_stopping_config: Optional[EarlyStoppingConfig] | None = None):

        new_task = create_finetune(api_key=self.api_key, name=name, dataset_id=dataset_id, model=model, n_epochs=n_epochs, batch_size=batch_size, wandb_key=wandb_key,
                        learning_rate=learning_rate,n_checkpoints_and_evaluations_per_epoch=n_checkpoints_and_evaluations_per_epoch,
                        save_only_best_checkpoint=save_only_best_checkpoint, train_with_lora=train_with_lora, lora_config=lora_config, early_stopping_config=early_stopping_config)
        
        print("New Task created successfully.")
        print(new_task)
        
        return new_task